import random

def file(n=100, pcommand=0.9, k=15):
    '\n'.join(command_or_comment(pcom, k) for _ in range(n))

def command_or_comment(pcommand, k):
    return command(k) if random.random() < pcommand else comment()

def command(k):
    args_part = ''.join(intersperse(args(), ['\n', ' ']))
    return ''.join([identifier(k), '(', args_part, ')'])

def intersperse(ls, seps):
    """
    intersperse(ls, seps) puts a separator between each element of
    ls, randomly chosen from seps.
    """
    return [y for x in ls for y in [x, random.choice(seps)]][:-1]

numbers = map(str, range(10))
low_letters = map(chr, range(ord('a'), ord('z') + 1))
high_letters = [c.upper() for c in low_letters]
identifier_chars = numbers + low_letters + high_letters + ['_']

def identifier(k):
    """
    identifier(k) generates an identifier with up to k letters.
    """
    return ''.join(random.choice(identifier_chars)
                   for _ in range(random.randint(1, k+1)))

def args():
    pass




from __future__ import print_function
import argparse
import sys

import cmakelists_parsing.parsing as cmp

def main():
    # Parse arguments
    parser = argparse.ArgumentParser(description='Pretty-print CMakeLists files.')
    parser.add_argument('files', type=str, nargs='*',
                        help='files to pretty print (default is stdin)')
    parser.add_argument('-t', '--tree', action='store_true',
                        help='print out the syntax trees')
    args = parser.parse_args()

    # Gather files
    filenames = args.files
    files = [('<stdin>', sys.stdin)]
    if filenames:
        files = [(name, open(name)) for name in filenames]

    # Process files
    for (name, file) in files:
        with file:
            input = file.read()
            tree = cmp.parse(input, path=name)
            if args.tree:
                # Print out AST
                print(repr(tree))
            else:
                # Pretty print
                print(str(tree), end='')

if __name__ == '__main__':
    main()




def merge_pairs(list, should_merge, merge):
    """
    Merges adjacent elements of list using the function merge
    if they satisfy the predicate should_merge.

    >>> merge_pairs([], None, None)
    []

    >>> merge_pairs([1], None, None)
    [1]

    >>> merge_pairs([1, 2], lambda x, y: False, None)
    [1, 2]

    >>> merge_pairs([1, 2], lambda x, y: y == x + 1, lambda x, y: (x, y))
    [(1, 2)]

    >>> merge_pairs([1, 2, 3], lambda x, y: y == x + 1, lambda x, y: (x, y))
    [(1, 2), 3]

    >>> merge_pairs([1, 2, 3, 4], lambda x, y: y == x + 1, lambda x, y: (x, y))
    [(1, 2), (3, 4)]
    """
    ret = []
    i = 0
    while i < len(list) - 1:
        a = list[i]
        b = list[i + 1]
        if should_merge(a, b):
            ret.append(merge(a, b))
            i += 2
        else:
            ret.append(a)
            i += 1
    if i == len(list) - 1:
        ret.append(list[i])
    return ret



import glob
import os
import unittest

import cmakelists_parsing.parsing as cmp

def yield_examples():
    paths = glob.glob(os.path.join(os.path.dirname(__file__),
                                   '..', 'example_inputs', '*'))
    for path in paths:
        with open(path) as file:
            contents = file.read()
            yield path, contents

class ExamplesTestCase(unittest.TestCase):
    def test_idempotency_of_parse_unparse(self):
        round_trip = lambda s, path='<string>': str(cmp.parse(s, path))
        for path, contents in yield_examples():
            print 'Testing on %s' % path
            self.assertEqual(round_trip(contents, path),
                             round_trip(round_trip(contents, path)),
                             'Failed on %s' % path)

    def test_tree_is_unchanged(self):
        for path, contents in yield_examples():
            expected = cmp.parse(contents, path)
            actual = cmp.parse(str(cmp.parse(contents, path)))
            msg = 'Failed on %s.\nExpected\n%s\n\nGot\n%s' % (path, expected, actual)
            self.assertEqual(expected, actual, msg)



from collections import namedtuple
import re

from cmakelists_parsing import list_utils

QuotedString = namedtuple('QuotedString', 'contents comments')
_Arg = namedtuple('Arg', 'contents comments')
_Command = namedtuple('Command', 'name body comment')
BlankLine = namedtuple('BlankLine', '')

class File(list):
    """Top node of the syntax tree for a CMakeLists file."""

    def __str__(self):
        '''
        Returns the pretty-print string for tree
        with indentation given by the string tab.
        '''
        return '\n'.join(compose_lines(self)) + '\n'

    def __repr__(self):
        return 'File(' + repr(list(self)) + ')'

class Comment(str):
    def __repr__(self):
        return 'Comment(' + str(self) + ')'

def Arg(contents, comments=None):
    return _Arg(contents, comments or [])

def Command(name, body, comment=None):
    return _Command(name, body, comment)

class CMakeParseError(Exception):
    pass

def prettify(s):
    """
    Returns the pretty-print of the contents of a CMakeLists file.
    """
    return str(parse(s))

def parse(s, path='<string>'):
    '''
    Parses a string s in CMakeLists format whose
    contents are assumed to have come from the
    file at the given path.
    '''
    nums_toks = tokenize(s)
    nums_items = list(parse_file(nums_toks))
    nums_items = attach_comments_to_commands(nums_items)
    items = [item for _, item in nums_items]
    return File(items)

def strip_blanks(tree):
    return File([x for x in tree if not isinstance(x, BlankLine)])

def compose_lines(tree):
    """
    Yields pretty-printed lines of a CMakeLists file.
    """
    tab = '\t'
    level = 0
    for item in tree:
        if isinstance(item, (Comment, str)):
            yield level * tab + item
        elif isinstance(item, BlankLine):
            yield ''
        elif isinstance(item, _Command):
            name = item.name.lower()
            if name in ('endfunction', 'endmacro', 'endif', 'else'):
                level -= 1
            for i, line in enumerate(command_to_lines(item)):
                offset = 1 if i > 0 else 0
                line2 = (level + offset) * tab + line
                yield line2

            if name in ('function', 'macro', 'if', 'else'):
                level += 1

# FIXME: Make this split into more lines if the result would be too wide.
def command_to_lines(cmd):
    sep = ' '
    final_paren = ')' if cmd.body and cmd.body[-1].comments else ')'
    comment_part = '  ' + cmd.comment if cmd.comment else ''
    result = cmd.name + '(' + sep.join(map(arg_to_str, cmd.body)) + final_paren + comment_part
    return [l.strip() for l in result.splitlines()]

def arg_to_str(arg):
    comment_part = '  ' + '\n'.join(arg.comments) + '\n' if arg.comments else ''
    return arg.contents + comment_part

def parse_file(toks):
    '''
    Yields line number ranges and top-level elements of the syntax tree for
    a CMakeLists file, given a generator of tokens from the file.

    toks must really be a generator, not a list, for this to work.
    '''
    prev_type = 'newline'
    for line_num, (typ, tok_contents) in toks:
        if typ == 'comment':
            yield ([line_num], Comment(tok_contents))
        elif typ == 'newline' and prev_type == 'newline':
            yield ([line_num], BlankLine())
        elif typ == 'word':
            line_nums, cmd = parse_command(line_num, tok_contents, toks)
            yield (line_nums, cmd)
        prev_type = typ

def attach_comments_to_commands(nodes):
    return list_utils.merge_pairs(nodes, command_then_comment, attach_comment_to_command)

def command_then_comment(a, b):
    line_nums_a, thing_a = a
    line_nums_b, thing_b = b
    return (isinstance(thing_a, _Command) and
            isinstance(thing_b, Comment) and
            set(line_nums_a).intersection(line_nums_b))

def attach_comment_to_command(lnums_command, lnums_comment):
    command_lines, command = lnums_command
    _, comment = lnums_comment
    return command_lines, Command(command.name, command.body[:], comment)

def parse_command(start_line_num, command_name, toks):
    cmd = Command(name=command_name, body=[], comment=None)
    expect('left paren', toks)
    for line_num, (typ, tok_contents) in toks:
        if typ == 'right paren':
            line_nums = range(start_line_num, line_num + 1)
            return line_nums, cmd
        elif typ == 'left paren':
            raise ValueError('Unexpected left paren at line %s' % line_num)
        elif typ in ('word', 'string'):
            cmd.body.append(Arg(tok_contents, []))
        elif typ == 'comment':
            c = tok_contents
            if cmd.body:
                cmd.body[-1].comments.append(c)
            else:
                cmd.comments.append(c)
    msg = 'File ended while processing command "%s" started at line %s' % (
        command_name, start_line_num)
    raise CMakeParseError(msg)

def expect(expected_type, toks):
    line_num, (typ, tok_contents) = next(toks)
    if typ != expected_type:
        msg = 'Expected a %s, but got "%s" at line %s' % (
            expected_type, tok_contents, line_num)
        raise CMakeParseError(msg)

# http://stackoverflow.com/questions/691148/pythonic-way-to-implement-a-tokenizer
# TODO: Handle multiline strings.
scanner = re.Scanner([
    (r'#.*',                lambda scanner, token: ("comment", token)),
    (r'"[^"]*"',            lambda scanner, token: ("string", token)),
    (r"\(",                 lambda scanner, token: ("left paren", token)),
    (r"\)",                 lambda scanner, token: ("right paren", token)),
    (r'[^ \t\r\n()#"]+',    lambda scanner, token: ("word", token)),
    (r'\n',                 lambda scanner, token: ("newline", token)),
    (r"\s+",                None), # skip other whitespace
])

def tokenize(s):
    """
    Yields pairs of the form (line_num, (token_type, token_contents))
    given a string containing the contents of a CMakeLists file.
    """
    toks, remainder = scanner.scan(s)
    if remainder != '':
        msg = 'Unrecognized tokens at line %s: %s' % (line_num, remainder)
        raise ValueError(msg)
    line_num = 1
    for tok_type, tok_contents in toks:
        yield line_num, (tok_type, tok_contents.strip())
        line_num += tok_contents.count('\n')




#!/usr/bin/env python

import unittest

from cmakelists_parsing.parsing import (
    File, Command, Comment, BlankLine, Arg, parse, prettify)

class ParsingTestCase(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_parse_empty_raises_exception(self):
        self.assertEqual(File([]), parse(''))

    def test_parse_nonempty1(self):
        input = 'FIND_PACKAGE(ITK REQUIRED)'
        output = parse(input)
        expected = File([Command('FIND_PACKAGE', [Arg('ITK'), Arg('REQUIRED')])])
        msg = '\nexpected\n%s\ngot\n%s' % (repr(expected), repr(output))
        self.assertEqual(expected, output, msg)

    def test_parse_nonempty2(self):
        input = '''\
# Top level comment
FIND_PACKAGE(ITK REQUIRED)
INCLUDE(${ITK_USE_FILE})

ADD_EXECUTABLE(CastImageFilter CastImageFilter.cxx)
TARGET_LINK_LIBRARIES(CastImageFilter # inline comment 1
vtkHybrid   #inline comment 2
ITKIO ITKBasicFilters ITKCommon
)
        '''

        output = parse(input)

        expected = File([
            Comment('# Top level comment'),
            Command('FIND_PACKAGE', [Arg('ITK'), Arg('REQUIRED')]),
            Command('INCLUDE', [Arg('${ITK_USE_FILE}')]),
            BlankLine(),
            Command('ADD_EXECUTABLE', [Arg('CastImageFilter'), Arg('CastImageFilter.cxx')]),
            Command('TARGET_LINK_LIBRARIES', [Arg('CastImageFilter', comments=['# inline comment 1']),
                                              Arg('vtkHybrid', comments=['#inline comment 2']),
                                              Arg('ITKIO'),
                                              Arg('ITKBasicFilters'),
                                              Arg('ITKCommon')]),
            ])
        msg = '\nexpected\n%s\ngot\n%s' % (expected, output)
        self.assertEqual(expected, output, msg)

    def test_idempotency_of_parsing_and_unparsing(self):
        input = '''\
# Top level comment
FIND_PACKAGE(ITK REQUIRED)
INCLUDE(${ITK_USE_FILE})
'''
        round_trip = lambda s: str(parse(s))
        self.assertEqual(round_trip(input), round_trip(round_trip(input)))

    def test_invalid_format_raises_an_exception(self):
        input = 'FIND_PACKAGE('
        self.assertRaises(Exception, parse, input)

    def test_line_numbers_in_exceptions(self):
        input = '''\
FIND_PACKAGE(ITK)
INCLUDE(
'''
        try:
            parse(input)
            self.fail('Expected an exception, but none was raised.')
        except Exception as e:
            self.assertTrue('line 2' in str(e))

    def test_arg_with_a_slash(self):
        tree = parse('include_directories (${HELLO_SOURCE_DIR}/Hello)')
        expected = File([
            Command('include_directories', [Arg('${HELLO_SOURCE_DIR}/Hello')])
            ])
        self.assertEqual(expected, tree)

    def test_command_with_no_args(self):
        tree = parse('cmd()')
        expected = File([Command('cmd', [])])
        self.assertEqual(expected, tree)

    def assertStringEqualIgnoreSpace(self, a, b):
        a2 = ''.join(a.split())
        b2 = ''.join(b.split())
        msg = '\nExpected\n%s\ngot\n%s\n(ignoring whitespace details)' % (a, b)
        self.assertEqual(a2, b2, msg)

    def test_arg_comments_preserved(self):
        input = '''
some_Command (
	x  # inline comment about x
	)
'''
        output = str(parse(input))

    def test_comments_preserved(self):
        input = '''\
# file comment
# more about the file

# comment above Command1
Command1(VERSION 2.6)  # inline comment for Command1

Command2(x  # inline comment about x
	"y"  # inline comment about a quoted string "y"
	)  # inline comment for Command2
'''
        output = str(parse(input))

        self.assertMultiLineEqual(input, output)

    def test_multiline_string(self):
        s = '''
string containing
newlines
'''
        input = '''\
set (MY_STRING "%s")
''' % s
        tree = parse(input)
        expected = File([Command('set', [Arg('MY_STRING'),
                                         Arg('"' + s + '"')])])
        self.assertEqual(expected, tree)

    # TODO: test macros and functions
    def test_ifs_indented(self):
        input = '''
if(a)
	if(b)
		set(X 1)
	endif()
else(a)
	if(c)
		set(Y 2)
	endif(c)
endif(a)
'''
        self.assertMultiLineEqual(input, str(parse(input)))

    def test_macros_indented(self):
        input = '''
macro(hello MESSAGE)
	message(${MESSAGE})
endmacro(hello)  # call the macro with the string "hello world"
hello("hello world")
'''
        self.assertUnchangedByPrettyPrinting(input)

    def test_functions_indented(self):
        input = '''
function(hello MESSAGE)
	message(${MESSAGE})
endfunction(hello)  # call the macro with the string "hello world"
hello("hello world")
'''
        self.assertUnchangedByPrettyPrinting(input)

    def assertUnchangedByPrettyPrinting(self, input):
        self.assertMultiLineEqual(input, prettify(input))

if __name__ == '__main__':
    unittest.main()




from setuptools import setup, find_packages

# Prevent "TypeError: 'NoneType' object is not callable" error
# when running python setup.py test 
# (see http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html)
try:
    import multiprocessing
except ImportError:
    pass

setup(
    name='cmakelists_parsing',
    version='0.3.1',
    author='Issac Trotts',
    author_email='issac.trotts@gmail.com',
    url='http://github.com/ijt/cmakelists_parsing',
    description='Parser for CMakeLists.txt files',
    packages=find_packages(),
    zip_safe=False,
    install_requires=['pyPEG2'],
    tests_require=['nose'],
    test_suite='nose.collector',
    include_package_data=True,
    entry_points = {
        'console_scripts': [
            'cmake_pprint = cmakelists_parsing.cmake_pprint:main',
        ]
    })

